from datetime import datetime
from typing import List, Optional
from pydantic import BaseModel
from uuid import UUID



class ItemModel(BaseModel):
    _id: UUID
    sequence: int
    id: UUID
    name: str
    title: str
    caption: str
    icon: str
    path: str
    route: str
    tags: Optional[str]


class MenuModel(BaseModel):
    _id: UUID
    node: str
    sequence: int
    id: UUID
    name: str
    title: str
    caption: str
    icon: str
    path: str
    route: str
    colour: str
    tags: Optional[str]
    items: Optional[List[ItemModel]]


class NavModel(BaseModel):
    node: str
    menus: List[MenuModel]
    
    
class AccessModel(BaseModel):
    navs: Optional[List[NavModel]]
    uix: dict


from typing import Any, Union
from uuid import UUID



def is_uuid(val: str, ver: int = 4) -> bool:
    try:
        return str(UUID(val, version=ver)) == val
    except ValueError:
        return False


def is_decimal(s):
    try:
        if isinstance(s, int):
            s = f"{s}.00"
        if isinstance(s, float):
            s = str(s)
        return '.' in s and s.count('.') == 1 
    except ValueError:
        return False


class SchemaField:
    def __init__(self, name: str, schema: dict):
        self.name = name
        self.typ = schema.get('typ')
        self.req = schema.get('req', True)
        self.min = schema.get('min', 0)
        self.max = schema.get('max', 0)
        self.len = schema.get('len', 0)
        self.iss = schema.get('iss', f"Field '{name}' is either missing or invalid.")
        self.prop: Union[dict[str, dict[str, Any]], None] = schema.get('prop')
        self.enum = schema.get('enum')

        if not self.typ or self.typ not in {'int', 'bool', 'str', 'uuid', 'float', 'decimal', 'list', 'dict'}:
            self.error = f"Schema field '{name}' has missing/invalid 'type' attribute."
        else:
            self.error = None


class Command:
    
    def __init__(self) -> None:
        self.errors = []
        self.is_valid = True

    def _validate(self, field: SchemaField, key: str, value: Any) -> None:
        hints = []
        val_str = str(value) if value is not None else ""

        if field.req and (value is None or val_str == ''):
            hints.append(f"Field '{key}' is required.")

        if field.min > 0 and len(val_str) < field.min:
            hints.append(f"Field '{key}' must have a minimum length of {field.min}.")

        if 0 < field.max < len(val_str):
            hints.append(f"Field '{key}' must have a maximum length of {field.max}.")

        if field.len > 0 and len(val_str) != field.len:
            hints.append(f"Field '{key}' must be of length {field.len}.")

        if field.typ == 'decimal':
            if not is_decimal(val_str):
                hints.append(f"Field '{key}' is not a valid decimal.")
        elif field.typ == 'uuid':
            if not is_uuid(val_str):
                hints.append(f"Field '{key}' is not a valid UUID.")
        elif value is not None and type(value).__name__ != field.typ:
            hints.append(f"Field '{key}' must be of type {field.typ}, {type(value).__name__} provided.")

        if field.enum and value not in field.enum:
            enum_str = ", ".join(map(str, field.enum))
            hints.append(f"Field '{key}' must be one of [{enum_str}].")

        if hints:
            self.errors.append({
                'type': "VALIDATION ERROR",
                'field': key,
                'error': field.iss,
                'hints': hints
            })

    def check(self, schema: dict[str, dict], data: dict) -> None:
        for key, field_schema in schema.items():
            field = SchemaField(name=key, schema=field_schema)

            if field.error:
                self.errors.append({
                    'type': "SCHEMA ERROR",
                    'field': key,
                    'error': field.error
                })
                continue

            value: Any = data.get(key)

            if field.typ == 'dict' and isinstance(value, dict) and field.prop:
                self.check(schema=field.prop, data=value)

            self._validate(field=field, key=key, value=value)

        if self.errors:
            self.is_valid = False




from abc import ABC, abstractmethod
from typing import Union



class BaseDataxMigration(ABC):
    
    @abstractmethod
    def migrate(self) -> Union[str, None]:
        pass
        
    @abstractmethod
    def reverse(self) -> Union[str, None]:
        pass
        



from typing import Optional
from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.models.nation_model import NationModel
from nkunyim_util.caches.nation_cache import NationCache
from nkunyim_util.api.nkunyim_api_client import NkunyimApiClient


from .signals import nation_data_updated


class NationService:

    def __init__(self, req: HttpRequest, session_key: str, code: str) -> None:
        self.cache = NationCache(key=f"nat.{session_key}")
        self.req = req
        self.code = code


    def _get_from_api(self) -> Optional[NationModel]:
        try:
            client = NkunyimApiClient(req=self.req, name=settings.PLACE_SERVICE)
            response = client.get(path=f"/api/nations/?code={self.code.upper()}")
            if not response.ok:
                # log error
                return None
            
            json_data = response.json()
            result_data = dict(json_data).get("data", [None])
            model_data = result_data[0]

            nation_model = NationModel(**model_data)
            self.cache.set(model=nation_model, timeout=60 * 60 * 24)
            
            # Inform interested parties
            nation_data_updated.send(sender=NationModel, instance=nation_model)
            
            return nation_model
        except:
            return None
            
    
    def get(self) -> Optional[NationModel]:
        return self.cache.get() or self._get_from_api()


import csv
from datetime import datetime
import logging
from pathlib import Path

from django.conf import settings

from .logging_command import LoggingCommand, LOG_TYPE_KEY, LOG_TYPE_GEN



class LoggingHandler(logging.Handler):

    def emit(self, record: logging.LogRecord) -> None:
        try:
            data = record.__dict__.copy()
            stream = data[LOG_TYPE_KEY] if LOG_TYPE_KEY in data else LOG_TYPE_GEN
            command = LoggingCommand()
            response = command.send(stream=stream, data=data)
            if not response.ok:
                self._write_to_csv(data)

        except Exception:
            self.handleError(record)

    def _write_to_csv(self, data: dict) -> None:
        date_str = datetime.now().strftime("%Y%m%d")
        file_path = Path(settings.TEXT_FILE_PATH) / f"{date_str}.csv"
        file_path.parent.mkdir(parents=True, exist_ok=True)

        write_header = not file_path.exists() or file_path.stat().st_size == 0

        with file_path.open(mode='a', newline='', encoding='utf-8') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=data.keys())

            if write_header:
                writer.writeheader()

            writer.writerow(data)



from django.dispatch import Signal


# Nation signal
nation_data_updated = Signal()


# Application signal
app_data_updated = Signal()

# User signal
user_data_updated = Signal()

# Access signal
access_data_updated = Signal()



from pydantic import BaseModel


class PageModel(BaseModel):
    uri: str
    host: str
    subdomain: str
    domain: str
    path: str
    dirs: list[str]
    node: str
    name: str
    root: str





from django.http import HttpRequest
from user_agents.parsers import UserAgent

from nkunyim_util.caches.user_agent_cache import UserAgentCache
from nkunyim_util.models.user_agent_model import UserAgentModel



class UserAgentService:
    
    def __init__(self, req: HttpRequest, session_key: str) -> None:
        self.cache = UserAgentCache(key=f"ua.{session_key}")
        self.session_key = session_key
        self.req = req
    
    def _make(self) -> UserAgentModel:
        ua_string = self.req.META.get('HTTP_USER_AGENT') or \
            self.req.headers.get('User-Agent') or \
            self.req.headers.get('user-agent', '')

        if isinstance(ua_string, bytes):
            ua_string = ua_string.decode('utf-8', 'ignore')

        ua = UserAgent(ua_string)

        data = {
            "is_mobile": ua.is_mobile,
            "is_tablet": ua.is_tablet,
            "is_touch_capable": ua.is_touch_capable,
            "is_pc": ua.is_pc,
            "is_bot": ua.is_bot,
            "is_email_client": ua.is_email_client,
            "browser_name": ua.browser.family,
            "browser_version": ua.browser.version_string,
            "os_name": ua.os.family,
            "os_version": ua.os.version_string,
            "device_name": ua.device.family,
            "device_brand": ua.device.brand,
            "device_model": ua.device.model,
        }
        
        user_agent = UserAgentModel(**data)
        self.cache.set(model=user_agent, timeout=60 * 60 * 24)
        
        return user_agent
    
    
    def get(self) -> UserAgentModel:
        return self.cache.get() or self._make()



from django.http import HttpRequest

from nkunyim_util.models.context_model import ContextModel
from nkunyim_util.services.access_service import AccessService
from nkunyim_util.services.app_service import AppService
from nkunyim_util.services.location_service import LocationService
from nkunyim_util.services.nation_service import NationService
from nkunyim_util.services.page_service import PageService
from nkunyim_util.services.user_agent_service import UserAgentService
from nkunyim_util.services.user_service import UserService


class ContextService:
    
    def __init__(self, req: HttpRequest) -> None:
        self.req = req
        
    
    def create(self) -> ContextModel:
        page_model = PageService(req=self.req)
        page_data = page_model.get()
        
        app_model = AppService(req=self.req, session_key=page_data.root)
        app_data = app_model.get()
        
        user_model = UserService()
        user_data = user_model.get(req=self.req)
    
        access_model = AccessService(req=self.req, session_key=page_data.root, application_id=app_data.id)
        access_data = access_model.get()
    
        location_model = LocationService(req=self.req, session_key=page_data.root)
        location_data = location_model.get()
        
        nation_model = NationService(req=self.req, session_key=page_data.root, code=location_data.country_code)
        nation_data = nation_model.get()
        
        user_agent_model = UserAgentService(req=self.req, session_key=page_data.root)
        user_agent_data = user_agent_model.get()
        
        data = {
            **app_data.__dict__,
            "page": page_data.__dict__,
            "user": user_data.__dict__,
            "uix": access_data.uix,
            "navs": access_data.navs,
            "role": access_data.role,
            "location": location_data.__dict__,
            "nation": nation_data.__dict__,
            "user_agent": user_agent_data.__dict__,
            "root": page_data.root
        }
        
        return ContextModel(**data)


from django.core.cache import DEFAULT_CACHE_ALIAS,  caches


DEFAULT_CACHE_TIMEOUT: int = 60 * 60 * 24


class CacheManager:
    """
    A simple cache manager to handle caching operations.
    """

    def __init__(self, cache_alias: str = DEFAULT_CACHE_ALIAS) -> None:
        self.cache = caches[cache_alias]


    def set(self, key, value, timeout=None):
        """
        Set a value in the cache.
        
        :param key: The key under which the value is stored.
        :param value: The value to store.
        :param timeout: The time in seconds before the cache expires.
        """
        if self.cache:
            self.cache.set(key, value, timeout)


    def get(self, key):
        """
        Get a value from the cache.
        
        :param key: The key of the cached value.
        :return: The cached value or None if not found.
        """
        if self.cache:
            return self.cache.get(key)
        return None


    def delete(self, key):
        """
        Delete a value from the cache.
        
        :param key: The key of the cached value to delete.
        """
        if self.cache:
            self.cache.delete(key)
            
            
    def clear(self):
        """
        Clear the entire cache.
        """
        if self.cache:
            self.cache.clear()
            
     


from datetime import datetime
from pydantic import BaseModel
from typing import List, Optional
from uuid import UUID


class AddressModel(BaseModel):
    id: UUID
    street_address: str
    postal_code: str
    locality: str
    city: str
    nation_id: UUID
    timezone: str
    language: str
    is_primary: bool
    is_verified: bool
    is_active: bool
    created_at: datetime
    updated_at: datetime


class RoleModel(BaseModel):
    id: UUID
    name: str
    title: str
    is_admin: str
    
    
class UserModel(BaseModel):
    id: UUID
    username: str
    nickname: str
    first_name: str
    middle_name: str
    last_name: str
    email_address: str
    email_verified: bool
    email_verified_at: datetime
    phone_number: str
    phone_verified: bool
    phone_verified_at: datetime
    gender: str
    birth_date: datetime
    photo_url: str
    photo_verified: bool
    photo_verified_at: datetime
    profile: str
    is_pwd: bool
    is_active: bool
    is_admin: bool
    is_verified: bool
    created_at: datetime
    updated_at: datetime
    addreses: Optional[List[AddressModel]]
    role: Optional[RoleModel]
    permissions: Optional[List[str]]



from django.contrib.gis.geoip2 import GeoIP2
from django.http import HttpRequest

from nkunyim_util.caches.location_cache import LocationCache
from nkunyim_util.models.location_model import LocationModel

 


class LocationService:
    
    FALLBACK_IP = "154.160.22.132"
    PRIVATE_IP_PREFIXES = ("192.168.",)
    LOCALHOST_SUFFIXES = (".0.0.1",)

    def __init__(self, req: HttpRequest, session_key: str) -> None:
        self.cache = LocationCache(key=f"loc.{session_key}")
        self.user_ip = self._extract_ip(req)
        self.req = req
    
    def _extract_ip(self, request: HttpRequest) -> str:
        ip = (
            request.META.get("HTTP_X_FORWARDED_FOR")
            or request.META.get("HTTP_X_REAL_IP")
            or request.META.get("REMOTE_ADDR", "")
        ).split(",")[0].strip()

        if ip.startswith(self.PRIVATE_IP_PREFIXES) or any(ip.endswith(suffix) for suffix in self.LOCALHOST_SUFFIXES):
            return self.FALLBACK_IP
        return ip
    
    def _make(self) -> LocationModel:
        geoip = GeoIP2()
        data = geoip.city(self.user_ip)
        data['user_ip'] = self.user_ip
        location_model = LocationModel(**data)
        self.cache.set(model=location_model, timeout=60 * 60 * 24)
        
        return location_model
        
    def get(self) -> LocationModel:
        return self.cache.get() or self._make()
    
    


from django.http import HttpRequest

from nkunyim_util.models.context_model import ContextModel
from nkunyim_util.services.context_service import ContextService


class Context:
    
    def __init__(self, req: HttpRequest) -> None:
        service = ContextService(req=req)
        self.model = service.create()
        
        
    def get(self) -> ContextModel:
        return self.model


from typing import Optional
from django.http import HttpRequest

from nkunyim_util.models.logging_model import LoggingModel
from nkunyim_util.services.location_service import LocationService
from nkunyim_util.services.page_service import PageService
from nkunyim_util.services.user_agent_service import UserAgentService


class LoggingService:
    
    def __init__(self, req: HttpRequest) -> None:
        self.req = req


    def get(self, xan: str, service_name: Optional[str] = None) -> LoggingModel:
        page_model = PageService(req=self.req)
        page_data = page_model.get()
        
        if not service_name:
            service_name = page_data.subdomain
            
        location_model = LocationService(req=self.req, session_key=page_data.root)
        location_data = location_model.get()
        
        user_agent_model = UserAgentService(req=self.req, session_key=page_data.root)
        user_agent_data = user_agent_model.get()
        
        logging_model = LoggingModel(
            xan=xan,
            service_name=service_name,
            location=location_data,
            user_agent=user_agent_data
        )
        
        return logging_model


from django.http import HttpRequest

from nkunyim_util.models.page_model import PageModel


class PageService:
    
    def __init__(self, req: HttpRequest) -> None:
        self.req = req
        
        
    def get(self) -> PageModel:      
        uri = self.req.build_absolute_uri()
        host = self.req.get_host()
        host_parts = host.lower().split('.')
        subdomain = host_parts[-3] if len(host_parts) > 2 else "www"
        domain = f"{host_parts[-2]}.{host_parts[-1]}"
        path = self.req.path.lower()
        dirs = path.strip('/').split('/') if path.strip('/') else ['/']
        node = dirs[-1] if dirs else "index"

        data = {
            "uri": uri,
            'host': host,
            'subdomain': subdomain,
            'domain': domain,
            'root': f"{subdomain}.{domain}",
            "path": path,
            "dirs": dirs,
            "node": node,
            "name": f"{node.title()}Page"
        }
        
        return PageModel(**data)


from .client import HttpClient, OidcClient
from .command import Command, BaseUserCommand, BaseAppCommand
from .encryption import Encryption
from .location import Location
from .messaging import Messaging
from .query import Query
from .session import HttpSession, OidcSession


from typing import Union
from django.core.mail import EmailMultiAlternatives


class EmailMessage(object):
    
    def compose(self, from_email: str, to_emails:list[str], subject: str, body: str) -> None:
        self.from_email = from_email
        self.to_emails = to_emails
        self.subject = subject
        self.body = body
        self.cc: Union[list[str], None] = None
        self.bcc: Union[list[str], None] = None
        self.attachments = []
        self.html = None
        
    def add_cc(self, cc_address: str) -> None:
        if not self.cc:
            self.cc = []
        self.cc.append(cc_address)
        
    def add_bc(self, bc_address: str) -> None:
        if not self.bcc:
            self.bcc = []
        self.bcc.append(bc_address)
        
    def add_html(self, html: str) -> None:
        self.html = html
        
    def add_attachment(self, attachment: str) -> None:
        self.attachments.append(attachment)
        
    def send(self) -> None:
        email = EmailMultiAlternatives(
            subject=self.subject,
            body=self.body,
            from_email=self.from_email,
            to=self.to_emails
        )
        if self.cc:
            email.cc = self.cc
            
        if self.attachments:
            email.attachments = self.attachments
            
        if self.bcc:
            email.bcc = self.bcc
            
        if self.html:
            email.attach_alternative(content=self.html, mimetype='text/html')
            
        email.send()